import torch,sys,pathlib
sys.path.append(pathlib.Path(__file__).parent.parent.as_posix())
from torchenhanced import CosineWarmup
import matplotlib.pyplot as plt

def test_cosine():
    pass #(for manual thingy only)
    # opti = torch.optim.SGD([torch.zeros(1)],lr=0.1)
    # sched = CosineWarmup(optimizer=opti, warmup_steps=10, T_0=100,
    #                      lr_shrink=0.8,lr_init=1e-3,T_mult=1,eta_min=1e-3)
    
    # lrs = []
    # epochs = []
    # for ep in range(600):
    #     # SImulate 600 epochs
    #     for batch in range(400):
    #         # Simulate batch
    #         sched.step(ep+batch/400)
    #         epochs.append(ep+batch/400)
    #         lrs.append(sched.get_last_lr())
    # # Plot lrs against epochs
    # plt.plot(epochs,lrs)
    # plt.show()

# test_cosine()


import torch.nn as nn
import torch


class DevModule(nn.Module):
    """
        Extremely small wrapper for nn.Module.
        Simply adds a method device() that returns
        the current device the module is on. Changes if
        self.to(...) is called.

        args :
        device : optional, default 'cpu'. Device to initialize the module on.
    """
    def __init__(self, device:str='cpu'):
        super().__init__()

        self.register_buffer('_devtens',torch.empty(0, device=device))

    @property
    def device(self):
        return self._devtens.device

    @property
    def paranum(self):
        return sum(p.numel() for p in self.parameters())


class ConfigModule(DevModule):
    """
        Same as DevModule, but with a config property that
        stores the necessary data to reconstruct the model.
        Use preferably over DevModule, especially with use with Trainer.

        args :
        config : Dictionary that contains the key:value pairs needed to 
        instantiate the model (i.e. the argument values of the __init__ method)
    """
    def __init__(self, config:dict, device:str='cpu'):
        super().__init__(device=device)

        self._config = config

        self.class_name = self.__class__.__name__ # Use this instead if, at time of saving, you need the name.
        # self._config['name'] = self.__class__.__name__ # Decided against using name in config, it defeats the purpose.

    @property
    def config(self):
        """
            Returns a json-serializable dict containing the config of the model.
            Essentially a key-value dictionary of the init arguments of the model.
            Should be redefined in sub-classes.
        """
        return self._config




from pathlib import Path
import json, os, torch


def load_string(file_name):
    try: return Path(file_name).read_text()
    except: return ''

def save_string(string, file_name):
    try: Path(file_name).write_text(string)
    except: return print(f"Error writing to {file_name=}")

def load_json(file_name): return json.loads(load_string(file_name))
def save_json(obj, file_name): save_string(json.dumps(obj), file_name)


def save_torch(obj, file_path, *, replace=True, **params): 
    if not os.path.isfile(file_path) or replace: return torch.save(obj, file_path, **params) # wrapper to unify notation
    else:
        print(f"File {file_path} already exists ! Use replace=True to overwrite it.")


import torch, torch.nn as nn, sys, pathlib, os
sys.path.append(pathlib.Path(__file__).parent.parent.as_posix())
import torch.optim.lr_scheduler as lrsched
from torch.utils.data import DataLoader
from src.torchenhanced import Trainer,ConfigModule

curfold = pathlib.Path(__file__).parent

class LinSimple(ConfigModule):
    def __init__(self, hidden = 1, out =1):
        config = locals()
        del config['self']
        del config['__class__']

        super().__init__(config)

        self.layer = nn.Linear(hidden, out)
    
    def forward(self, x):
        return self.layer(x)

class SimpleDataset(torch.utils.data.Dataset):
    def __init__(self):
        self.data = torch.arange(0,100).reshape((100,1)).float()
        self.targets = torch.arange(1,101).reshape((100,1)).float()
    
    def __getitem__(self, index):
        # print('Dispensing item nb: ',index)
        return self.data[index],self.targets[index] # (1,1), (1,1)
    
    def __len__(self):
        return len(self.data)

class LinearTrainer(Trainer):
    def __init__(self, run_name: str = None, project_name: str = None, state_save_loc=None,reach_plateau=100,run_config={}):
        model = LinSimple()
        opti = torch.optim.Adam(model.parameters(),lr=1e-3)
        schedo = lrsched.LinearLR(opti,start_factor=0.01,end_factor=1,total_iters=reach_plateau)

        super().__init__(model,optim=opti,scheduler=schedo, run_name=run_name, 
                         project_name=project_name,state_save_loc=state_save_loc,run_config=run_config)

        self.dataset =SimpleDataset() # Numbers from 0 to 99
        self.loss_val = []
        
    def get_loaders(self, batch_size, num_workers=0):
        return DataLoader(self.dataset, batch_size=batch_size, shuffle=False),None
    
    def process_batch(self, batch_data,**kwargs):
        x, y = batch_data # (B,1), (B,1)
        print(f'Processing batch {self.batches} : {x}')
        pred = self.model(x) # (B,1)
        loss = ((pred-y)**2).mean() # MSE

        if(self.do_batch_log):
            self.logger.log({'lossme/train':loss.item()},commit=False)
            self.logger.log({'other/lr':self.scheduler.get_last_lr()[0]},commit=False)
        
        return loss
    
    def process_batch_valid(self, batch_data):
        x, y = batch_data # (B,1), (B,1)
        
        pred = self.model(x) # (B,1)
        loss = ((pred-y)**2).mean() # MSE

        self.loss_val.append(loss.item())

        return loss
    
    def valid_log(self):
        self.logger.log({'lossme/valid':sum(self.loss_val)/len(self.loss_val)},commit=False)
        self.loss_val = []
        

def test_resume_train():
    """
        Testing is not automatic... Gotta look at the output, whether it does resume at the correct batch
    """
    trainer = LinearTrainer(run_name='test_saveforresume', project_name='test_torchenhanced', state_save_loc=os.path.join(curfold),reach_plateau=7)
    print('-----------------INITIAL TRAINING-----------------')
    trainer.train_steps(steps=29, batch_size=1, step_log=1,batch_tqdm=False, save_every=5,aggregate=2, valid_every=1)
    # print(f'Finished at batches : {trainer.batches}, epochs : {trainer.epochs}, steps : {trainer.steps_done}')
    
    # Test without resuming
    trainer3 = LinearTrainer(run_name='test_without_resume', project_name='test_torchenhanced', state_save_loc=os.path.join(curfold),reach_plateau=7)
    print('-----------------WITHOUT RESUMING TRAINING-----------------')
    trainer3.load_state(os.path.join(curfold,'test_torchenhanced','state','test_saveforresume.state')) # Load the state
    trainer3.train_steps(steps=5, batch_size=1, step_log=1, batch_tqdm=False,save_every=5,aggregate=2, valid_every=1,resume_batches=False,pickup=False)

    # Test with resuming
    trainer2 = LinearTrainer(run_name='test_with_resume', project_name='test_torchenhanced', state_save_loc=os.path.join(curfold),reach_plateau=7)
    print('-----------------RESUMING TRAINING-----------------')
    trainer2.load_state(os.path.join(curfold,'test_torchenhanced','state','test_saveforresume.state')) # Load the state
    trainer2.train_steps(steps=29, batch_size=1, step_log=1, batch_tqdm=False,save_every=5,aggregate=2, valid_every=1,resume_batches=True,pickup=True)


if __name__ == "__main__":
    test_resume_train()



import torch.optim.lr_scheduler as lrsched
from torch.optim.optimizer import Optimizer
import math

class _enable_get_lr_call:

    def __init__(self, o):
        self.o = o

    def __enter__(self):
        self.o._get_lr_called_within_step = True
        return self

    def __exit__(self, type, value, traceback):
        self.o._get_lr_called_within_step = False

class CosineWarmup(lrsched.CosineAnnealingWarmRestarts):
    """
        LR-scheduler with cosine learning rate, and a warmup phase additionally.
        Either input values for number of BATCHES, and .step() each batch, or input
        values for EPOCHS, and .step(epoch+batch/total_batches) each batch. This is less
        than ideal in my opinion, but the Trainer integrates that behavior (for now)

        Parameters :
        optimizer : torch.optim.Optimizer
            Optimizer on which the learning rate will act.
        warmup_steps : int
            Number of warmup steps.
        T_0 : int
            Cosine (initial) period in number of steps
        lr_init : float
            Value of lr at start of warmup
        lr_shrink : float (0.,1.)
            Multiplies lr at each reset, shrinking it
        T_mult : float (1.,infty)
            How much the period is multiplied at each restart (every period)
        eta_min : float
            minimum lr that is reached during oscillations
    """

    def __init__(self, optimizer: Optimizer, warmup_steps: int,T_0:int,lr_shrink:float=0.75, lr_init:float=1e-7,T_mult=1, eta_min=0, last_epoch=-1, verbose=False) -> None:
        self.warmup_steps=warmup_steps
        self.t=last_epoch
        self.lr_init = lr_init
        self.lr_shrink = lr_shrink
        self.lr_shrink_i = 1
        super().__init__(optimizer, T_0, T_mult, eta_min, last_epoch, verbose)

        

    def get_lr(self):
        if(self.t<=self.warmup_steps):
            # Warming up
            return [self.lr_init+(base_lr-self.lr_init)*self.t/self.warmup_steps for base_lr in self.base_lrs]
        else :
            # It's warm. T_cur should have correct translated value.
            return [self.lr_shrink_i*(self.eta_min + (base_lr - self.eta_min) * (1 + math.cos(math.pi * self.T_cur / self.T_i)) / 2)
                for base_lr in self.base_lrs]

    def step(self, epoch = None):
        """
            Almost equivalent to original cosineAnnealing step, but with warmup.
        """
        if epoch is None and self.last_epoch < 0:
            epoch = 0

        if epoch is None:
            epoch = self.last_epoch + 1
        
        if epoch < 0:
            raise ValueError("Expected non-negative epoch, but got {}".format(epoch))
    
        # Update values using closed-form
        self.t=epoch
        if self.t>self.warmup_steps:
            t_eff = self.t-self.warmup_steps
            if t_eff >= self.T_0:
                if self.T_mult == 1:
                    n = t_eff//self.T_i
                    self.T_cur = t_eff % self.T_0
                    self.T_i=self.T_0
                else:
                    n = int(math.log((t_eff / self.T_0 * (self.T_mult - 1) + 1), self.T_mult))
                    self.T_cur = t_eff - self.T_0 * (self.T_mult ** n - 1) / (self.T_mult - 1)
                    self.T_i = self.T_0 * self.T_mult ** (n)
                self.lr_shrink_i = self.lr_shrink**n
            else:
                self.T_i = self.T_0
                self.t=epoch
                self.T_cur = self.t-self.warmup_steps
    
        self.last_epoch = math.floor(epoch)



        with _enable_get_lr_call(self):
            for i, data in enumerate(zip(self.optimizer.param_groups, self.get_lr())):
                param_group, lr = data
                param_group['lr'] = lr
                self.print_lr(self.verbose, i, lr, epoch)

        self._last_lr = [group['lr'] for group in self.optimizer.param_groups]



import torch, torch.nn as nn, sys, pathlib
sys.path.append(pathlib.Path(__file__).parent.parent.as_posix())

import torch.optim.lr_scheduler as lrsched
from torch.optim import Optimizer
from torch.utils.data import DataLoader,Subset
# Import mnist for tests
from torchvision.datasets import MNIST
import torch.nn.functional as F
from torchvision import transforms as t
from src.torchenhanced import Trainer, DevModule, ConfigModule
import  os,time

curfold = pathlib.Path(__file__).parent

class LinSimple(ConfigModule):
    def __init__(self, hidden = 28*28, out =10):
        config = locals()
        del config['self']
        del config['__class__']

        super().__init__(config)

        self.layer = nn.Linear(hidden, out)
    
    def forward(self, x):
        return self.layer(x)
    
class LinearTrainer(Trainer):
    def __init__(self, run_name: str = None, project_name: str = None, state_save_loc=None,reach_plateau=100,run_config={},parallel=None,device='cpu'):
        model = LinSimple()
        opti = torch.optim.Adam(model.parameters(),lr=1e-3)
        schedo = lrsched.LinearLR(opti,start_factor=0.01,end_factor=1,total_iters=reach_plateau)

        super().__init__(model,optim=opti,scheduler=schedo, run_name=run_name, 
                         project_name=project_name,state_save_loc=state_save_loc,run_config=run_config,parallel=parallel,device=device)

        self.dataset =Subset(MNIST(os.path.join(curfold,'data'),download=True,transform=t.ToTensor()),range(100))
    
    def get_loaders(self, batch_size, num_workers=0):
        self.loss_val = []
        return DataLoader(self.dataset, batch_size=batch_size, shuffle=False), DataLoader(self.dataset, batch_size=batch_size, shuffle=True)
    
    def process_batch(self, batch_data,**kwargs):
        x, y = batch_data
        x = x.to(self.device)
        y = y.to(self.device)

        x = x.reshape((x.shape[0],-1))
        
        pred = self.model(x) # (B,10)
        loss = F.cross_entropy(pred,y,reduction='mean') 

        # assert self.stepnum==data_dict['stepnum'], f"Stepnum mismatch {self.stepnum} vs {data_dict['stepnum']}"
        # assert self.step_log==data_dict['step_log'], f"Step_log mismatch {self.step_log} vs {data_dict['step_log']}"
        # assert self.epoch == data_dict['epoch'], f"Epoch mismatch {self.epoch} vs {data_dict['epoch']}"
        # assert self.batchnum == data_dict['batchnum'], f"Batchnum mismatch {self.batchnum} vs {data_dict['batchnum']}"

        if(self.do_batch_log):
            self.logger.log({'lossme/train':loss.item()},commit=False)
            self.logger.log({'other/lr':self.scheduler.get_last_lr()[0]},commit=False)
        
        return loss
    
    def process_batch_valid(self, batch_data):
        x, y = batch_data
        x = x.to(self.device)
        y = y.to(self.device)
        x = x.reshape((x.shape[0],-1))
        
        pred = self.model(x) # (B,10)
        loss = F.cross_entropy(pred,y,reduction='mean') 

        self.loss_val.append(loss.item())

        return loss
    
    def valid_log(self):
        self.logger.log({'lossme/valid':sum(self.loss_val)/len(self.loss_val)},commit=False)
        self.loss_val = []
        

# FOR MANUAL TESTING, COULDN'T FIGURE OUT HOW TO AUTOMATE IT
# EPOCHS :
trainer = LinearTrainer(run_name='test_epochs', project_name='test_torchenhanced', 
                        state_save_loc=os.path.join(curfold),reach_plateau=200, run_config={'manamajeff':True},parallel=[0],device='cpu')
trainer.change_lr(1e-4)
if(os.path.exists(os.path.join(curfold,'test_torchenhanced','state','test_epochs.state'))):
    trainer.load_state(os.path.join(curfold,'test_torchenhanced','state','test_epochs.state'))
time.sleep(2)

trainer.train_epochs(epochs=100, batch_size=10, step_log=30, save_every=60,aggregate=2,batch_sched=True)

#STEPS :
trainer = LinearTrainer(run_name='test_steps', project_name='test_torchenhanced', state_save_loc=os.path.join(curfold),reach_plateau=1500,parallel=[0],device='cpu')

trainer.change_lr(1e-4)
if(os.path.exists(os.path.join(curfold,'test_torchenhanced','state','test_steps.state'))):
    trainer.load_state(os.path.join(curfold,'test_torchenhanced','state','test_steps.state'))

time.sleep(2)
trainer.train_steps(steps=1000, batch_size=10, step_log=30, save_every=60,aggregate=2, valid_every=100)


def test_resume_train():
    trainer = LinearTrainer(run_name='test_resume_train', project_name='test_torchenhanced', state_save_loc=os.path.join(curfold),reach_plateau=1500)
    trainer.change_lr(1e-4)
    trainer.train_steps(steps=1000, batch_size=200, step_log=30, save_every=60,aggregate=2, valid_every=100)
    trainer.save_state()
    print(f'Finished at batches : {trainer.batches}, epochs : {trainer.epochs}, steps : {trainer.steps_done}')
    
    # Test with resuming
    trainer2 = LinearTrainer(run_name='test_with_resume', project_name='test_torchenhanced', state_save_loc=os.path.join(curfold),reach_plateau=1500)
    trainer2.change_lr(1e-4)
    trainer2.load_state(os.path.join(curfold,'test_torchenhanced','state','test_resume_train.state')) # Load the state
    trainer2.train_steps(steps=200, batch_size=10, step_log=30, save_every=3000,aggregate=2, valid_every=100, resume_batches=True)

    # Test without resuming
    trainer3 = LinearTrainer(run_name='test_without_resume', project_name='test_torchenhanced', state_save_loc=os.path.join(curfold),reach_plateau=1500)
    trainer3.change_lr(1e-4)
    trainer3.load_state(os.path.join(curfold,'test_torchenhanced','state','test_resume_train.state')) # Load the state
    trainer3.train_steps(steps=200, batch_size=10, step_log=30, save_every=3000,aggregate=2, valid_every=100)

    assert trainer2.batches == trainer3.batches, f"Batch mismatch : {trainer2.batches} vs {trainer.batches}"
    assert trainer2.epochs == trainer3.epochs, f"Epoch mismatch : {trainer2.epochs} vs {trainer.epochs}"
    assert trainer2.steps_done == trainer3.steps_done, f"Step mismatch : {trainer2.steps_done} vs {trainer.steps_done}"

    # Look on wandb to see if they are the same


def test_Save_Weights():
    lintra = LinearTrainer(run_name='test_save_weights', project_name='AnewDawn', state_save_loc=os.path.join(curfold))
    lintra.save_state()

    Trainer.save_model_from_state(state_path=os.path.join(curfold,'AnewDawn','state','test_save_weights.state'),
                                 save_dir=os.path.join(curfold,'AnewDawn'),name='testJEFF')

    assert os.path.isfile(os.path.join(curfold,'AnewDawn','testJEFF.pt')), "Weights not found"
    assert os.path.isfile(os.path.join(curfold,'AnewDawn','testJEFF.config')), "Config not found"

def test_Trainer_config():
    ma = LinSimple(hidden=32,out=15)
    config = ma.config
    assert config == {'hidden':32, 'out':15, 'name':'LinSimple'}, f"Invalid config : {config}"


# Probably need to add more unit_tests...



import torch
from math import prod

def coord_to_flat(coord_tens,dimensions) :
    """
        Given a coordinate tensor of size (...,N), which contains
        the coordinates of points in N dimensions, we return a tensor of size (...,1),
        which contains the same coordinates, for the spacetime flattened from N to 1 dimensions.

        Parameters :
        coord_tens : torch.Tensor (..., N)
            The tensor containing the coordinates of an N-dimensional space
        dimensions : N-tuple
            The size of each of the dimensions. For instance, for an image of size H*W, dimensions = (H,W)
        
        Returns : 
        tensor of size (...), with the flattened indices.

        Example :
        For coord_tens of size (3) where T=[1,2,3], representing a point in 3d space (1,2,3), and given dimensions
        3,3,3, it would output 1*3*3+3*2+3=15, which is the location of the coordinate (1,2,3) in the flattened tensor.
        The extra dimensions are just treated as batch dimensions.
    """
    ndims = coord_tens.shape[-1]
    assert (len(dimensions)==ndims)

    flat_index_tensor=torch.zeros_like(coord_tens[...,0])

    for dimnum in range(ndims):
        flat_index_tensor+= coord_tens[...,dimnum]*prod(dimensions[dimnum+1:])
    
    return flat_index_tensor



## Includes some function to manipulate image tensors
import torch


def rgb_to_yuv(img) :
    """
        Convert a (B,3,...) RGB [0.,1.] tensor to a (B,3,...) [0,1.] YCbCr tensor.
    """
    img_yuv =torch.zeros_like(img)
    (R,G,B)  = img[:,0]*255,img[:,1]*255,img[:,2]*255
    img_yuv[:,0]=    + 0.299*R +.587*G+.114*B
    img_yuv[:,1]= 128- 0.169*R -0.331 *G+0.5*B
    img_yuv[:,2]= 128+ 0.5  *R -0.419 *G-0.081*B

    return img_yuv/255.

def yuv_to_rgb(img):
    """
        Converts a (B,3,...) YCbCr [0,255] tensor to a (B,3,...) RGB [0.,1.] tensor.
    """
    img_rgb = torch.zeros_like(img)
    (Y,Cb,Cr)  = img[:,0]*255,img[:,1]*255.-128,img[:,2]*255.-128

    img_rgb[:,0]= Y+0*Cb +1.4*Cr
    img_rgb[:,1]= Y-0.343 *Cb-0.711*Cr
    img_rgb[:,2]= Y+1.765 *Cb-0*Cr
    
    return img_rgb/255.

def convert_to_nbit(imgs,n):
    """
        Convert a (B,3,...) 8-bit-channel tensor with values in 0.-1.,
        to a (B,1,...) tensor with nbit-encoded RGB values. For now, n should
        be divisible by 3.
    """
    dtype=torch.int32

    if(n%3!=0):
        raise ValueError(f"n={n} not divisible by 3")

    chn = n//3
    imgs=(imgs*255.).to(dtype=dtype)
    
    # Introduce 0's in front (keep only the topmost chanbits bits)
    R=(imgs[:,0]>>(8-chn))
    G=(imgs[:,1]>>(8-chn))
    B=(imgs[:,2]>>(8-chn))

    img_coded=(R<<chn*2) | (G<<chn) | B

    #Restore 1-dim channel :

    return img_coded[:,None]


def decode_from_nbit(imgnbit,n):
    """
        Given a (B,1,...) n-bit coded tensor (hence with int values,
        in the range 0 to 2^n-1), decodes to a standard (B,3,...) tensor
        with 8-bit coded channels, and values in floats [0.,1.].
    """
    if(n%3!=0):
        raise ValueError(f"n={n} not divisible by 3")
    
    chn=n//3
    dtype=torch.int32

    trailing = torch.tensor(2**(8-chn-1),dtype=dtype) # Trailing bits which where lost. We choose around the middle

    # Erasing the trailing bits, and replaces them with "trailing"
    # We "and" with 255 (0xFF), to keep numbers only up to that value
    R = (((imgnbit >> chn*2) << (8-chn)) & 0xFF) | trailing
    G = ((imgnbit >> chn) << (8-chn)) & 0xFF | trailing
    B = (imgnbit << (8-chn)) & 0xFF | trailing

    return torch.cat([R,G,B],dim=1)/255.



import torch,sys,pathlib
sys.path.append(pathlib.Path(__file__).parent.parent.as_posix())

from src.torchenhanced.util import *


def test_coord_to_flat_batch():
    unfolded = torch.randint(0,100,(3,4,5,6)) #4-dim coord tensor
    folded = unfolded.reshape((-1))

    dimensions = unfolded.shape
    testcoord = torch.tensor([[1,2,3,4],[0,2,3,2]]) # Set of coordinate points

    flatcoord = coord_to_flat(testcoord,dimensions)
    assert flatcoord.shape == (2,)
    res = folded[flatcoord]
    for i in range(testcoord.shape[0]):
        assert (res[i]-unfolded[testcoord[i][0]][testcoord[i][1]][testcoord[i][2]][testcoord[i][3]]).all()==0

def test_coord_to_flat_unbatch():
    unfolded = torch.randint(0,100,(3,4,5,6)) #4-dim coord tensor
    folded = unfolded.reshape((-1))

    dimensions = unfolded.shape
    testcoord = torch.tensor([1,2,3,4]) # Set of coordinate points

    flatcoord = coord_to_flat(testcoord,dimensions)
    assert flatcoord.shape==(), f"Invalid shape : {flatcoord.shape}"
    res = folded[flatcoord]
    
    assert (res-unfolded[testcoord[0]][testcoord[1]][testcoord[2]][testcoord[3]]).all()==0

def test_color_comp_incomplete():
    from PIL import Image
    from torchvision import transforms as t

    # ntar=12
    # imgex=torch.clamp(torch.zeros((1,3,1,1)),0,1)
    # imgex[:,0]=140
    # imgex[:,1]=23
    # imgex[:,2]=23
    # print("==================================================")
    # # print("imgex shape : ",imgex.shape)
    # print("Initial : ",(imgex).squeeze())
    # imgex=convert_to_nbit(imgex/255.,ntar)
    # print("After : ",imgex.squeeze())
    # imgex=decode_from_nbit(imgex,ntar)
    # print("Decoded again : ",imgex.squeeze())


    # imgex= Image.open("Riffelsee.jpg")
    # imgex=t.ToTensor()(imgex)[None]
    # showTens(imgex)
    # imgex=rgb_to_yuv(imgex)
    
    
    # imgn = convert_to_nbit(imgex,ntar)
    # imgn = decode_from_nbit(imgn,ntar)


    # print("shape : ",imgex.shape)

    # showTens(yuv_to_rgb(imgn))

    # imgyuv = yuv_to_rgb(rgb_to_yuv(imgex))
    # print(f"BOUNDS : (maxy : {torch.max(imgyuv[:,0])},maxCb= {torch.max(imgyuv[:,1])}, maxCr= {torch.max(imgyuv[:,2])})")


import matplotlib.pyplot as plt
import os, torch, torchvision.transforms as transf
from torchvision.utils import make_grid
import cv2, numpy as np

@torch.no_grad()
def showTens(tensor, columns=None) :
    '''
        shows tensor as an image. Accepts (H,W), (C,H,W) and (*,C,H,W).
    '''
    if(len(tensor.shape)==2):
        fig = plt.figure()
        plt.imshow(tensor[None,:,:])
        plt.show()
    elif(len(tensor.shape)==3) :
        fig = plt.figure()
        plt.imshow(tensor.permute((1,2,0)))
        plt.show()
    elif(len(tensor.shape)==4) :
        # Assume B,C,H,W
        B=tensor.shape[0]
        if(columns is not None):
            numCol=columns
        else :
            numCol=min(8,B)

        fig = plt.figure()
        
        to_show=make_grid(tensor,nrow=numCol,pad_value=0.2 ,padding=3)
        if(tensor.shape[1]==1):
            to_show=to_show.mean(dim=0,keepdim=True)

        plt.imshow(to_show.permute(1,2,0))
        if(tensor.shape[1]==1):
            plt.colorbar()
        plt.axis('off')
        plt.show()
    elif(len(tensor.shape)>4):
        tensor = tensor.reshape((-1,tensor.shape[-3],tensor.shape[-2],tensor.shape[-1])) # assume all batch dimensions
        print("WARNING : assuming extra dimension are all batch dimensions, newshape : ",tensor.shape)
        showTens(tensor,columns)
    else :
        raise Exception(f"Tensor shape should be (H,W), (C,H,W) or (*,C,H,W), but got : {tensor.shape} !")

@torch.no_grad()
def saveTensVideo(tensor,folderpath,name="videotensor",columns=None,fps=30,out_size=800):
    """
        Saves tensor as a video. Accepts both (T,H,W), (T,3,H,W) and (*,T,3,H,W).

        Args:
        tensor : (T,H,W) or (T,3,H,W) or (*,T,3,H,W) video tensor
        folderpath : path to save the video
        name : name of the video
        columns : number of columns to use for the grid of videos (default 8 or less)
        fps : fps of the video (default 30)
        out_size : Width of output video (height adapts to not deform videos) (default 800)
    """
    if(len(tensor.shape)<3):
        raise ValueError(f"Tensor shape should be (T,H,W), (T,3,H,W) or (*,T,3,H,W), but got : {tensor.shape} !")
    elif(len(tensor.shape)==3):
        # add channel dimension
        tensor=tensor[:,None,:,:].expand(-1,3,-1,-1) # (T,3,H,W)
        saveTensVideo(tensor,folderpath,name,columns)
    elif(len(tensor.shape)==4):
        if(tensor.shape[1]==1):
            print('Assuming gray-scale video')
            tensor=tensor.expand(-1,3,-1,-1) # (T,3,H,W)
        assert tensor.shape[1]==3, f"Tensor shape should be (T,H,W), (T,3,H,W) or (*,T,3,H,W), but got : {tensor.shape} !"
        # A single video
        _make_save_video(tensor,folderpath,name,fps)
    elif(len(tensor.shape)==5):
        if(tensor.shape[2]==1):
            print('Assuming gray-scale video')
            tensor=tensor.expand(-1,-1,3,-1,-1)
        assert tensor.shape[2]==3, f"Tensor shape should be (T,H,W), (T,3,H,W) or (*,T,3,H,W), but got : {tensor.shape} !"
        # Assume B,T,3,H,W
        B,T,C,H,W = tensor.shape

        if(columns is not None):
            numCol=columns
        else :
            numCol=min(8,B)


        black_cols = (-B)%numCol
        video_tens = torch.cat([tensor.to('cpu'),torch.zeros(black_cols,T,C,H,W)],dim=0) # (B',T,3,H,W)
        video_tens = transf.Pad(3)(video_tens) # (B',T,3,H+3*2,W+3*2)

        B,T,C,H,W = video_tens.shape
        resize_ratio = out_size/(H*numCol)
        indiv_vid_size = int(H*resize_ratio),int(W*resize_ratio)

        video_tens = video_tens.reshape((B*T,C,H,W)) # (B'*T,3,H,W
        video_tens = transf.Resize(indiv_vid_size,antialias=True)(video_tens) # (B'*T,3,H',W')
        video_tens = video_tens.reshape((B,T,C,indiv_vid_size[0],indiv_vid_size[1])) # (B',T,3,H',W')
        B,T,C,H,W = video_tens.shape

        assert B%numCol==0
        numRows = B//numCol

        video_tens = video_tens.reshape((numRows,numCol,T,C,H,W)) # (numRows,numCol,T,3,H',W')
        video_tens = torch.einsum('nmtchw->tcnhmw',video_tens) # (T,C,numRows,H',numCol,W')
        video_tens = video_tens.reshape((T,C,numRows*H,numCol*W)) # (T,C,numRows*H,numCol*W)

        _make_save_video(video_tens,folderpath,name,fps)
    elif (len(tensor.shape)>5):
        video_tens = tensor.reshape((-1,*tensor.shape[-4:]))
        saveTensVideo(video_tens,folderpath,name,columns,fps,out_size)


def gridify(tensor,out_size=800,columns=None):
    """
        Makes a grid of images/videos from a batch of images. Accepts (B,*,H,W), 

        Args:
            tensor : (B,*,H,W) tensor
            out_size : size of the output grid
            columns : number of columns of the grid
        
        Returns:
            (*,H',W') tensor, representing the grid of images/videos
    """
    B,H,W = tensor.shape[0],tensor.shape[-2],tensor.shape[-1]
    device = tensor.device
    if(columns is not None):
        numCol=columns
    else :
        numCol=min(8,B)


    black_cols = (-B)%numCol
    tensor = torch.cat([tensor,torch.zeros(black_cols,*tensor.shape[1:],device=device)],dim=0) # (B',*,H,W)
    tensor = transf.Pad(3)(tensor) # (B',*,H+3*2,W+3*2)

    B,H,W = tensor.shape[0],tensor.shape[-2],tensor.shape[-1]
    rest_dim = tensor.shape[1:-2]

    rest_dim_prod = 1
    for dim in rest_dim:
        rest_dim_prod*=dim
    
    resize_ratio = out_size/(H*numCol)

    indiv_vid_size = int(H*resize_ratio),int(W*resize_ratio)
    tensor = tensor.reshape((B,rest_dim_prod,H,W))
    tensor = transf.Resize(indiv_vid_size,antialias=True)(tensor) # (B',rest_dim_prod,H',W')
    B,H,W = tensor.shape[0],tensor.shape[-2],tensor.shape[-1]

    assert B%numCol==0

    numRows = B//numCol

    tensor = tensor.reshape((numRows,numCol,rest_dim_prod,H,W)) # (numRows,numCol,rest_dim_prod,H',W')
    tensor = torch.einsum('nmrhw->rnhmw',tensor) # (rest_prod,numRows,H',numCol,W')
    tensor = tensor.reshape((rest_dim_prod,numRows*H,numCol*W)) # (rest_prod,numRows*H,numCol*W)
    tensor = tensor.reshape((*rest_dim,numRows*H,numCol*W)) # (*,numRows*H,numCol*W)

    return tensor

@torch.no_grad()
def saveTensImage(tensor, folderpath,name="imagetensor",columns=None):
    '''
        Saves tensor as an image. Accepts both (C,H,W) and (*,C,H,W). 
    '''
    if(len(tensor.shape)==2) :
        fig = plt.figure()
        plt.imshow(tensor[None,:,:])
        plt.savefig(os.path.join(folderpath,f"{name}.png"))
    if(len(tensor.shape)==3) :
        fig = plt.figure()
        plt.imshow(tensor.permute((1,2,0)))
        plt.savefig(os.path.join(folderpath,f"{name}.png"))
    elif(len(tensor.shape)==4) :
        # Assume B,C,H,W
        B=tensor.shape[0]
        if(columns is not None):
            numCol=columns
        else :
            numCol=min(8,B)

        fig = plt.figure()
        
        to_show=make_grid(tensor,nrow=numCol,pad_value=0. ,padding=2)
        if(tensor.shape[1]==1):
            to_show=to_show.mean(dim=0,keepdim=True)

        plt.imshow(to_show.permute(1,2,0))
        if(tensor.shape[1]==1):
            plt.colorbar()
        #createGrid(tensor,fig,numCol)
        plt.axis('off')
        plt.savefig(os.path.join(folderpath,f"{name}.png"))
    elif(len(tensor.shape)>4):
        tensor = tensor.reshape((-1,tensor.shape[-3],tensor.shape[-2],tensor.shape[-1])) # assume all batch dimensions
        print("WARNING : assuming extra dimension are all batch dimensions, newshape : ",tensor.shape)
        saveTensImage(tensor,folderpath,name,columns)
    else :
        raise Exception(f"Tensor shape should be (H,W), (C,H,W) or (*,C,H,W), but got : {tensor.shape} !")

def _make_save_video(video_tens,folderpath,name,fps=30):
    """
        Makes a video in mp4 and saves it at the given folderpath, with given name.

        Args:
        video_tens : (T,C,H,W) tensor
        path : path to save the video
    """
    T,C,H,W = video_tens.shape
    output_file = os.path.join(folderpath,f"{name}.mp4")
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    video = cv2.VideoWriter(output_file, fourcc, fps, (W, H))
        
    to_save = (255*video_tens.permute(0,2,3,1).cpu().numpy()).astype(np.uint8)

    for t in range(T):
        frame = to_save[t]
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        video.write(frame)

    video.release()


import torch.nn as nn, math
import torch, wandb, os
import torch.optim.lr_scheduler as lrsched
from torch.optim import Optimizer
from datetime import datetime
from tqdm import tqdm
from .modules import DevModule, ConfigModule


class Trainer(DevModule):
    """
        Mother class used to train models, exposing a host of useful functions.
        Should be subclassed to be used, and the following methods should be redefined :
            - process_batch, mandatory
            - get_loaders, mandatory
            - epoch_log, optional
            - valid_log, optional
            - process_batch_valid, mandatory if validation is used (i.e. get_loaders returns 2 loaders)
        For logging, use wandb.log, which is already initialized. One should be logged in into the wandb
        account to make the logging work. See wandb documentation for info on logging.
        
        Use train_epochs OR train_steps, according to whether you would like to train at epoch level or at batch number level.
        Loading a state trained with train_epochs and using it in train_steps will cause unexpected behavior, and vice-versa.

        Parameters :
        model : Model to be trained
        optim : Optimizer to be used. ! Must be initialized
        with the model parameters ! Default : AdamW with 1e-3 lr.
        scheduler : Scheduler to be used. Can be provided only if using
        non-default optimizer. Must be initialized with aforementioned 
        optimizer. Default : warmup for 4 epochs from 1e-6.
        state_save_loc : str or None(default), folder in which to store data 
        pertaining to training, such as the training state, wandb folder and model weights.
        device : torch.device, device on which to train the model
        parallel : None or list[int,str], if None, no parallelization, if list, list of devices (int or torch.device) to parallelize on
        run_name : str, for wandb and saves, name of the training session
        project_name : str, name of the project in which the run belongs
        run_config : dict, dictionary of hyperparameters (any). Will be viewable in wandb.
    """

    def __init__(self, model : nn.Module, optim :Optimizer =None, scheduler : lrsched._LRScheduler =None,*, 
                 state_save_loc=None,device:str ='cpu',parallel:list[int]=None, run_name :str = None,project_name :str = None,
                 run_config : dict = {}):
        super().__init__()
        
        self.parallel_train = parallel is not None
        self.parallel_devices = parallel
        if(self.parallel_train):
            # Go to GPU if parallel training
            device = self.parallel_devices[0]
        self.to(device)
        self.model = model.to(device)

        if(state_save_loc is None) :
            self.data_fold = os.path.join('.',project_name)
            self.state_save_loc = os.path.join(self.data_fold,"state")
            self.model_save_loc = os.path.join(self.data_fold,"weights")
        else :
            self.data_fold = os.path.join(state_save_loc,project_name)#
            
            self.state_save_loc = os.path.join(state_save_loc,project_name,"state")
            self.model_save_loc = os.path.join(state_save_loc,project_name,"weights")
        
        os.makedirs(self.data_fold,exist_ok=True)
        if(optim is None):
            self.optim = torch.optim.AdamW(self.model.parameters(),lr=1e-3)
        else :
            self.optim = optim

        if(scheduler is None):
            self.scheduler = lrsched.LinearLR(self.optim,start_factor=0.05,total_iters=4)
        else :
            self.scheduler = scheduler
        

        # Session hash, the date to not overwrite sessions
        self.session_hash = datetime.now().strftime('%H-%M_%d_%m')
        if(run_name is None):
            self.run_name = self.session_hash
            run_name= os.path.join('.','runs',self.session_hash)
        else :
            self.run_name=run_name
            run_name = os.path.join('.','runs',run_name)
 
        self.run_config = dict(model=self.model.__class__.__name__,
                            **run_config)

        self.run_id = wandb.util.generate_id() # For restoring the run
        self.project_name = project_name
        
        # Universal attributes for logging purposes
        self.stepnum = 0 # number of steps in current training instance
        self.batchnum = None # same as stepnum, DEPRECATED

        self.batches = 0 # number of total batches ever
        self.steps_done = 0 # number of total steps ever TODO (now its same as batches, will become optimizer steps later)
        self.epochs = 0 # number of total epochs ever
        self.samples = 0 # number of total samples ever

        self.step_log = None # number of steps between each log
        self.totbatch = None # total number of batches in one epoch for this training instance
        self.do_batch_log = False 

        # Used for logging instead of wandb.log, useful if wandb not imported
        self.logger = None
    
    def change_lr(self, new_lr):
        """
            Changes the learning rate of the optimizer.
            Might clash with scheduler ?
        """
        for g in self.optim.param_groups:
            g['lr'] = new_lr
        

    def load_state(self,state_path : str, strict: bool=True):
        """
            Loads Trainer state, for restoring a run.

            params : 
            state_path : location of the sought-out state_dict
            strict: whether to load the state_dict in strict mode or not
        """

        if(isinstance(self.model, nn.DataParallel)):
            # Unwrap the model, since we saved the state_dict of the model, not the DataParallel
            self.model = self.model.module.to(self.device)
        
        if(not os.path.exists(state_path)):
            raise ValueError(f'Path {state_path} not found, can\'t load state')

        state_dict = torch.load(state_path,map_location=self.device)
        if(self.model.config != state_dict['model_config']):
            print('WARNING ! Loaded model configuration and state model_config\
                  do not match. This may generate errors.')
            
        assert self.model.class_name == state_dict['model_name'], f'Loaded model {state_dict["model_name"]} mismatch with current: {self.model.class_name}!'
        assert self.optim.__class__.__name__ == state_dict['optim_name'], f'Loaded optimizer : {state_dict["optim_name"]} mismatch with current: {self.optim.__class__.__name__} !'
        assert self.scheduler.__class__.__name__ == state_dict['scheduler_name'], f'Loaded scheduler : {state_dict["scheduler_name"]} mismatch with current: {self.optim.__class__.__name__} !'

        self.model.load_state_dict(state_dict['model_state'],strict=strict)
        self.session_hash = state_dict['session']
        self.optim.load_state_dict(state_dict['optim_state'])
        self.scheduler.load_state_dict(state_dict['scheduler_state'])
        self.run_id = state_dict['run_id']
        self.steps_done = state_dict.get('steps_done',0)
        self.batches = state_dict.get('batches',0)

        self.epochs = state_dict.get('epochs',0)
        self.samples = state_dict.get('samples',0)
        self.run_config = state_dict.get('run_config',{'model':self.model.__class__.__name__})
        # Maybe I need to load also the run_name, we'll see

        print('Training state load successful !')
        print(f'Loaded state had {state_dict["epochs"]} epochs trained.')

    def load_model_from_state(self,state_path : str, strict: bool=True):
        """
            Loads only model weights from state. Useful if you just want to load a 
            pretrained model to train it on a different dataset.
        """
        if(isinstance(self.model, nn.DataParallel)):
            # Unwrap the model, since we saved the state_dict of the model, not the DataParallel
            self.model = self.model.module.to(self.device)
        
        if(not os.path.exists(state_path)):
            raise ValueError(f'Path {state_path} not found, can\'t load state')

        state_dict = torch.load(state_path,map_location=self.device)
        if(self.model.config != state_dict['model_config']):
            print('WARNING ! Loaded model configuration and state model_config\
                  do not match. This may generate errors.')
            
        assert self.model.class_name == state_dict['model_name'], f'Loaded model {state_dict["model_name"]} mismatch with current: {self.model.class_name}!'

        self.model.load_state_dict(state_dict['model_state'],strict=strict)

        print('Model load successful !')
        print(f'Loaded model had {state_dict["epochs"]} epochs trained.')

    def save_state(self,epoch:int = None):
        """
            Saves trainer state. Describe by the following dictionary :

            state_dict : dict, contains at least the following key-values:
                - 'model' : contains model.state_dict
                - 'session' : contains self.session_hash
                - 'optim' :optimizer
                - 'scheduler : scheduler
                - 'model_config' : json allowing one to reconstruct the model.
                - 'run_id' : id of the run, for wandb
                - 'steps_done' : only applicable in case of step training, number of steps done
                - 'samples' : number of samples seen
            If you want a more complicated state, training_epoch should be overriden.

            Args :
            epoch : int, if not None, will append the epoch number to the state name.
        """
        os.makedirs(self.state_save_loc,exist_ok=True)
        
        # Avoid saving the DataParallel
        saving_model = self.model.module if self.parallel_train else self.model
    
        # Create the state
        try :
            model_config = saving_model.config
        except AttributeError as e:
            raise AttributeError(f"Error while fetching model config ! Make sure model.config is defined. (see ConfigModule doc).")

        state = dict(
        optim_state=self.optim.state_dict(),scheduler_state=self.scheduler.state_dict(),model_state=saving_model.state_dict(),
        model_name=saving_model.class_name,optim_name=self.optim.__class__.__name__,scheduler_name=self.scheduler.__class__.__name__,
        model_config=model_config,session=self.session_hash,run_id=self.run_id, steps_done=self.steps_done,epochs=self.epochs,
        samples=self.samples, batches=self.batches,run_config=self.run_config
        )

        name = self.run_name
        if (epoch is not None):
            os.makedirs(os.path.join(self.state_save_loc,'backups'),exist_ok=True)
            name=os.path.join('backups',name+'_'+f'{epoch:.2f}')

        name = name + '.state'
        saveloc = os.path.join(self.state_save_loc,name)
        torch.save(state,saveloc)

        print(f'Saved training state at {datetime.now().strftime("%H-%M_%d_%m")}')
        print(f'At save, {self.epochs} epochs are done.')


    @staticmethod
    def save_model_from_state(state_path:str,save_dir:str='.',name:str=None):
        """
            Extract model weights and configuration, and saves two files in the specified directory,
            the weights (.pt) and a .config file containing the model configuration, which can be loaded
            as a dictionary with torch.load.

            Args :
            state_path : path to the trainer state
            save_dir : directory in which to save the model
            name : name of the model, if None, will be model_name_date.pt
        """
        namu, config, weights = Trainer.model_config_from_state(state_path,device='cpu')

        if (name is None):
            name=f"{namu}_{datetime.now().strftime('%H-%M_%d_%m')}"
        name=name+'.pt'
        os.makedirs(save_dir,exist_ok=True)
        saveloc = os.path.join(save_dir,name)
        
        torch.save(weights, saveloc)

        torch.save(config, os.path.join(save_dir,name[:-3]+'.config'))

        print(f'Saved weights of {name} at {save_dir}/{name}  !')

    @staticmethod
    def opti_names_from_state(state_path: str,device='cpu'):
        """
            Given the path to a trainer state, returns a 2-tuple (opti_config, scheduler_config),
            where each config is a tuple of the name of the optimizer, and its state_dict.
            Usually useful only if you forgot which optimizer you used, but load_state should
            be used instead usually.
            
            Args :
            state_path : path of the saved trainer state
            device : device on which to load state

            Returns :
            2-uple, (optim_config, scheduler_config), where *_config = (name, state_dict)

            Example of use :
            get name from opti_config[0]. Use it with eval (or hardcoded) to get the class,
            instanciante : 
            optim = torch.optim.AdamW(model.parameters(),lr=1e-3)
            optim.load_state_dict(opti_config[1])
        """
        if(not os.path.exists(state_path)):
            raise ValueError(f'Path {state_path} not found, can\'t load config from it')

        state_dict = torch.load(state_path,map_location=device)
        opti_name = state_dict['optim_name']
        opti_state = state_dict['optim_state']
        sched_name = state_dict['sched_name']
        sched_state = state_dict['sched_state']

        return (opti_name,opti_state),(sched_name,sched_state)

    @staticmethod
    def config_from_state(state_path: str,device: str=None):
        """
            Given the path to a trainer state, returns a tuple (config, weights)
            for the saved model. The model can then be initialized by using config 
            as its __init__ arguments, and load the state_dict from weights.

            Args :
            state_path : path of the saved trainer state
            device : device on which to load. Default one if None specified

            returns: 3-uple
            model_name : str, the saved model class name
            config : dict, the saved model config (instanciate with element_name(**config))
            state_dict : torch.state_dict, the model's state_dict (load with .load_state_dict(weights))

        """
        print('WARNING : Deprecated, will be removed in next version')
        print('For model config, use model_config_from_state instead')
        if(not os.path.exists(state_path)):
            raise ValueError(f'Path {state_path} not found, can\'t load config from it')
        
        if(device is None):
            state_dict = torch.load(state_path)
        else :
            state_dict = torch.load(state_path,map_location=device)

        config = state_dict['model_config']
        model_name = state_dict['model_name']
        weights = state_dict['model_state']

        return model_name,config,weights

    @staticmethod
    def model_config_from_state(state_path: str,device: str=None):
        """
            Given the path to a trainer state, returns a tuple (config, weights)
            for the saved model. The model can then be initialized by using config 
            as its __init__ arguments, and load the state_dict from weights.

            Args :
            state_path : path of the saved trainer state
            device : device on which to load. Default one if None specified

            returns: 3-uple
            model_name : str, the saved model class name
            config : dict, the saved model config (instanciate with element_name(**config))
            state_dict : torch.state_dict, the model's state_dict (load with .load_state_dict(weights))

        """
        if(not os.path.exists(state_path)):
            raise ValueError(f'Path {state_path} not found, can\'t load config from it')
        
        if(device is None):
            state_dict = torch.load(state_path)
        else :
            state_dict = torch.load(state_path,map_location=device)

        config = state_dict['model_config']
        model_name = state_dict['model_name']
        weights = state_dict['model_state']

        return model_name,config,weights
    
    @staticmethod
    def run_config_from_state(state_path: str,device: str=None):
        """
            Given the path to a trainer state, returns the run_config dictionary.

            Args :
            state_path : path of the saved trainer state
            device : device on which to load. Default one if None specified

            returns: dict, the run_config dictionary
        """
        if(not os.path.exists(state_path)):
            raise ValueError(f'Path {state_path} not found, can\'t load config from it')
        
        if(device is None):
            state_dict = torch.load(state_path)
        else :
            state_dict = torch.load(state_path,map_location=device)

        return state_dict['run_config']

    def process_batch(self,batch_data,**kwargs):
        """
            Redefine this in sub-classes. Should return the loss, as well as 
            the data_dict (potentially updated). Batch_data will be on 'cpu' most of the
            time, except if you dataset sets a specific device. Can do logging and other things 
            optionally. Loss is automatically logged, so no need to worry about it. 
            Use self.model to access the model.

            Args :
            batch_data : whatever is returned by the dataloader
            Default class attributes, automatically maintained by the trainer, are :
                - self.device : current model device
                - self.batchnum : current validation mini-batch number
                - self.step_log : number of steps (minibatches) interval in 
                        which we should log. (PREFER USING do_batch_log instead)
                - self.do_batch_log : whether we should log this batch or not
                - self.totbatch : total number of validation minibatches.
                - self.epoch : current epoch
                - self.samples : number of samples seen
            Returns : 2-uple, (loss, data_dict)
        """
        raise NotImplementedError('process_batch should be implemented in Trainer sub-class')

    def process_batch_valid(self,batch_data, **kwargs):
        """
            Redefine this in sub-classes. Should return the loss, as well as 
            the data_dict (potentially updated). Use self.model to access the model (it is already in eval mode).
            Batch_data will be on 'cpu' most of the time, except if you dataset sets a specific device. 
            There should be NO logging done inside this function, only in valid_log.
            Proper use should be to collect the data to be logged in a class attribute,
            and then log it in valid_log (to log once per epoch). Loss is automatically 
            logged, so no need to worry about it. 

            Args :
            batch_data : whatever is returned by the dataloader
            Default class attributes, automatically maintained by the trainer, are :
                - self.device : current model device
                - self.batchnum : current validation mini-batch number
                - self.step_log : number of steps (minibatches) interval in which we should log 
                    Minibatch logging in valid is not recommended, since it is not synchronized with the epoch x-axis.
                - self.totbatch : total number of validation minibatches.
                - self.epoch : current epoch
                - self.samples : number of samples seen

            Returns : 2-uple, (loss, data_dict)
        """
        raise NotImplementedError('process_batch_valid should be implemented in Trainer sub-class')

    def get_loaders(self,batch_size, num_workers=0):
        """
            Builds the dataloader needed for training and validation.
            Should be re-implemented in subclass.

            Args :
            batch_size

            Returns :
            2-uple, (trainloader, validloader)
        """
        raise NotImplementedError('get_loaders should be redefined in Trainer sub-class')

    def epoch_log(self):
        """
            To be (optionally) implemented in sub-class. Does the logging 
            at the epoch level, is called every epoch. Only log using commit=False,
            because of sync issues with the epoch x-axis.

            Args :
            data_dict : DEPRECATED ! Avoid using it. Use class attributes instead. 
            Default class attributes, automatically maintained by the trainer, are :
                - self.batchnum : current validation mini-batch number
                - self.step_log : number of steps (minibatches) interval in which we should log 
                - self.totbatch : total number of validation minibatches.
                - self.epoch : current epoch
                - self.samples : number of samples seen
        """
        pass

    def valid_log(self):
        """
            To be (optionally) implemented in sub-class. Does the logging 
            at the epoch level, is called every epoch. Only log using commit=False,
            because of sync issues with the epoch x-axis.


            Args :
            data_dict : DEPRECATED ! Avoid using it. Use class attributes instead. 
                Default class attributes, automatically maintained by the trainer, are :
                    - self.batchnum : current validation mini-batch number
                    - self.step_log : number of steps (minibatches) interval in which we should log 
                    - self.totbatch : total number of validation minibatches.
                    - self.epoch : current epoch
                    
                    - self.samples : number of samples seen

        """
        pass
    
    def train_init(self,**kwargs):
        """
            Can be redefined for doing stuff just at the beginning of the training,
            for example, freezing weights, preparing some extra variables, or anything really.
            Not mandatory, it is called at the very beginnig of train_epochs/train_steps. The
            dictionary 'train_init_params' is passed as parameter. As such, it can take
            any combination of parameters.
        """
        pass

    def train_epochs(self,epochs : int,batch_size:int,*,batch_sched:bool=False,save_every:int=50,
                     backup_every: int=None,step_log:int=None,
                     num_workers:int=0,aggregate:int=1,
                     batch_tqdm:bool=True,train_init_params:dict={}):
        """
            Trains for specified epoch number. This method trains the model in a basic way,
            and does very basic logging. At the minimum, it requires process_batch and 
            process_batch_valid to be overriden, and other logging methods are optionals.

            data_dict can be used to carry info from one batch to another inside the same epoch,
            and can be used by process_batch* functions for logging of advanced quantities.
            Params :
            epochs : number of epochs to train for
            batch_size : batch size
            batch_sched : if True, scheduler steps (by a lower amount) between each batch.
            Not that this use is deprecated, so it is recommended to keep False. For now, 
            necessary for some Pytorch schedulers (cosine annealing).
            save_every : saves trainer state every 'save_every' epochs
            step_log : If not none, will also log every step_log minibatches, in addition to each epoch
            batch_log : same as step_log, DEPRECATED
            num_workers : number of workers in dataloader
            aggregate : how many batches to aggregate (effective batch_size is aggreg*batch_size)
            load_from : path to a trainer state_dict. Loads the state
                of the trainer from file, then continues training the specified
                number of epochs.
            train_init_params : Parameter dictionary passed as argument to train_init
        """
        
        # Initiate logging
        self.logger = wandb.init(name=self.run_name,project=self.project_name,config=self.run_config,
                   id = self.run_id,resume='allow',dir=self.data_fold)
        
        # Define the custom x axis metric, epochs
        self.logger.define_metric("batches",hidden=True)
        self.logger.define_metric("epochs",hidden=True)
        self.logger.define_metric("ksamples",hidden=True)
        # For all plots, we plot against the epoch by default
        self.logger.define_metric("*", step_metric='epochs')

        self.train_init(**train_init_params)
        
        train_loader,valid_loader = self.get_loaders(batch_size,num_workers=num_workers)
        validate = valid_loader is not None

        self.model.train()
        if(self.parallel_train):
            print('Parallel training on devices : ',self.parallel_devices)
            self.model = nn.DataParallel(self.model,device_ids=self.parallel_devices)
        
        if(batch_sched):
            assert self.epochs-self.scheduler.last_epoch<1e-5, f'Epoch mismatch {self.epochs} vs {self.scheduler.last_epoch}'
        else:
            assert int(self.epochs)==self.scheduler.last_epoch, f'Epoch mismatch {self.epochs} vs {self.scheduler.last_epoch}'
        #Floor frac epochs, since we start at start of epoch, and also for the scheduler :
        self.epochs = int(self.epochs)
        print('Number of batches/epoch : ',len(train_loader))
        self.stepnum = 0 # This is the current instance number of steps, using for when to log save etc

        self.step_log = step_log
        step_loss=[]

        
        for ep_incr in tqdm(range(epochs)):
            epoch_loss,n_aggreg=[[],0]
            
            self.totbatch = len(train_loader)

            # Iterate with or without tqdm
            if(batch_tqdm):
                iter_on=tqdm(enumerate(train_loader),total=self.totbatch)
            else :
                iter_on=enumerate(train_loader)

            # Epoch of Training
            for batchnum,batch_data in iter_on :
                # Process the batch
                self.batchnum=batchnum
                epoch_loss, step_loss, n_aggreg = self._step_batch(batch_data,True,epoch_loss,step_loss,n_aggreg, aggregate, step_sched=False)
    
                if(batch_sched):
                    self.scheduler.step(self.epochs)

                self.stepnum+=1
                self.steps_done+=1
                self.batches+=1
                self.samples+=batch_size
                self.epochs+=1/self.totbatch

            self.epochs = round(self.epochs) # round to integer, should already be, but to remove floating point stuff
            
            if(not batch_sched):
                self.scheduler.step()
            else :
                self.scheduler.step(self.epochs)
            
            # Epoch of validation
            if(validate):
                self._validate(valid_loader,batch_tqdm)
                self.model.train()
                self.valid_log()
    

            # Log training at epoch level
            self.logger.log({'loss/train_epoch':sum(epoch_loss)/len(epoch_loss)},commit=False)
            self.epoch_log()
                
            self._update_x_axis(epoch_mode=True)
            
            # Save and backup when applicable
            self._save_and_backup(curstep=ep_incr,save_every=save_every,backup_every=backup_every)

        self.logger.finish()

    def train_steps(self,steps : int,batch_size:int,*,save_every:int=50,
                    backup_every: int=None, valid_every:int=1000,step_log:int=None,
                    num_workers:int=0,aggregate:int=1,pickup:bool=True,resume_batches:bool=False,
                    batch_tqdm:bool=True, train_init_params:dict={}):
        """
            Trains for specified number of steps(batches). This method trains the model in a basic way,
            and does very basic logging. At the minimum, it requires process_batch and 
            process_batch_valid to be overriden, and other logging methods are optionals. Epoch_log is not
            used in step level training.
            Note that the scheduler will be called AFTER EVERY MINIBATCH, i.e. after every step. Everything
            is logged by default against the number of steps, but the 'epochs' metric is also defined, and
            it depends on the size of the dataloader defined in get_loaders.

            Params :
            batch_size : batch size
            steps : number of steps (batches) to train for
            save_every : saves trainer state every 'save_every' epochs
            backup_every : saves trainer state without overwrite every 'backup_every' steps
            valid_every : validates the model every 'valid_every' steps
            step_log : If not none, used for logging every step_log steps. In process_batch,
            use self.do_step_log to know when to log. NOTE : if using aggregating, for step logging,
            we consider a step to be aggregate batches, not 'true' batches.
            num_workers : number of workers in dataloader
            aggregate : how many batches to aggregate (effective batch_size is aggreg*batch_size)
            pickup : if False, will train for exactly 'steps' steps. If True, will restart at the previous
            number of steps, and train until total number of steps is 'steps'. Useful for resuming training,
            if you want to train for a certain specific number of steps. In both cases, the training resumes
            where it left off, only difference is how many MORE steps it will do.
            resume_batches : if True, will resume training assuming the first self.batches on the dataloader
            are already done. Usually, use ONLY if dataloader does NOT shuffle.
            batch_tqdm : whether to use tqdm for the batch loop or not
            train_init_params : Parameter dictionary passed as argument to train_init
        """
    
        # Initiate logging
        self.logger = wandb.init(name=self.run_name,project=self.project_name,config=self.run_config,
                   id = self.run_id,resume='allow',dir=self.data_fold)
        
        # Define the custom x axis metrics
        self.logger.define_metric("epochs",hidden=True)
        self.logger.define_metric("batches",hidden=True)
        # For all plots, we plot against the batches by default, since we do step training
        self.logger.define_metric("*", step_metric='batches')

        self.train_init(**train_init_params)
        
        train_loader,valid_loader = self.get_loaders(batch_size,num_workers=num_workers)
        validate = valid_loader is not None

        self.totbatch = len(train_loader) # Number of batches in one epoch

        self.model.train()

        if(self.parallel_train):
            print('Parallel training on devices : ',self.parallel_devices)
            self.model = nn.DataParallel(self.model,device_ids=self.parallel_devices)
        
        print('Number of batches/epoch : ',len(train_loader)/1000 ,'k')

        self.step_log = step_log
        step_loss=[]

        steps_completed = False
        if(pickup):
            self.stepnum = self.batches
        else:
            self.stepnum = 0 #Current instance stepnumber, used for when to log and stop training.

        while not steps_completed:
            iter_on=enumerate(train_loader)

            if(resume_batches):
                resume_batches=False # Only resume for the first epoch, not if we reach and and restart.
                tofastforward = (self.batches)%self.totbatch
                print(f'Fast forwarding {self.batches}%{self.totbatch}={tofastforward} batches')
                for _ in tqdm(range(tofastforward)):
                    # skip batches already done
                    next(iter_on)
                if(batch_tqdm):
                    iter_on=tqdm(iter_on,total=self.totbatch-tofastforward)
            else :
                if(batch_tqdm):
                    iter_on=tqdm(iter_on,total=self.totbatch)
    
            n_aggreg=0
            # Epoch of Training
            for batchnum,batch_data in iter_on :
                # Process the batch according to the model.
                self.batchnum=batchnum
                _, step_loss, n_aggreg = self._step_batch(batch_data,False,[],step_loss,n_aggreg, aggregate,step_sched=True)

                # Validation if applicable
                if(validate and self.batches%valid_every==valid_every-1):
                    self._validate(valid_loader,batch_tqdm)
                    self.valid_log()
                    self._update_x_axis(epoch_mode=False)
                    self.model.train()


                self.stepnum+=1
                self.steps_done+=1 # TO BE MODIFIED TO BE OPTIMIZER STEPS
                self.batches+=1
                self.samples+=batch_size
                self.epochs +=1/self.totbatch

                # TODO minor bug, when we resume we shift by one minibatch the saving schedule
                # Comes because the first save location is at %valid_every-1, so it last one less step since we start stepnum at 0
    	        # Save and backup on the basis of BATCHES, not steps. That way when we resume, everything is consistent
                self._save_and_backup(self.batches,save_every,backup_every)


            
                if(self.stepnum>=steps):
                    steps_completed=True
                    self._save_and_backup(1,save_every,backup_every)
                    break
            
            

        wandb.finish()

    def _update_x_axis(self,epoch_mode):
        """
            Adds and commits pending wandb.log calls, and adds the x-axis metrics,
            to use the correct defaults.

            Args:   
            epoch_mode : bool, whether default x-axis is epoch or not
        """
        # TODO remove the epoch_mode, and just use epochs as last one, it doesn't make a difference anyway
        # TODO make so that batches area actually batches, and steps_done are optimizer steps
        self.logger.log({'ksamples' : self.samples//1000},commit=False)

        if(epoch_mode):
            self.logger.log({'batches': self.steps_done},commit=False)
            self.logger.log({'epochs': self.epochs},commit=True)
        else :
            self.logger.log({'epochs': self.epochs},commit=False)
            self.logger.log({'batches': self.steps_done},commit=True)

    def _step_batch(self, batch_data, epoch_mode, epoch_loss, step_loss, n_aggreg, aggregate, step_sched):
        """
            Internal function, makes one step of training given minibatch
        """
        n_aggreg+=1
        self.do_batch_log = self.stepnum%(self.step_log*aggregate)==0 #if aggregating, one 'batch' is actually aggregate batches

        loss = self.process_batch(batch_data)
        
        epoch_loss.append(loss.item())
        step_loss.append(loss.item())

        loss=loss/aggregate # Rescale loss if aggregating.
        loss.backward() # Accumulate gradients
        
        if(self.do_batch_log):
            self.logger.log({'loss/train_step':sum(step_loss)/len(step_loss)},commit=False)
            self._update_x_axis(epoch_mode=epoch_mode)
            step_loss=[]
            self.do_batch_log=False

        if(n_aggreg%aggregate==0):
            n_aggreg=0
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), 1.)

            self.optim.step()
            self.optim.zero_grad()
            if(step_sched):
                self.scheduler.step()
        
        return epoch_loss,step_loss,n_aggreg

    @torch.no_grad()
    def _validate(self,valid_loader, batch_tqdm)->None:
        self.model.eval()
        val_loss=[]
        t_totbatch = self.totbatch
        t_batchnum = self.batchnum

        self.totbatch = len(valid_loader) # For now we use same totbatch for train and valid, might wanna change that in the future
        if(batch_tqdm):
            print('------ Validation ------')
            iter_on=tqdm(enumerate(valid_loader),total=self.totbatch)
        else:
            iter_on=enumerate(valid_loader)

        for (v_batchnum,v_batch_data) in iter_on:
            self.batchnum=v_batchnum
            
            loss = self.process_batch_valid(v_batch_data)
            val_loss.append(loss.item())
        
        self.totbatch=t_totbatch
        self.batchnum=t_batchnum

        # Log validation data
        self.logger.log({'loss/valid':sum(val_loss)/len(val_loss)},commit=False)
    
    def _save_and_backup(self,curstep,save_every,backup_every):
        # We use curstep-1, to save at a moment consistent with the valid
        # And valid looks at curstep-1. (we updated curstep in between)
        if (curstep-1)%save_every==0 :
            self.save_state()
        
        if backup_every is not None:
            if (curstep-1)%backup_every==backup_every-1 :
                self.save_state(epoch=self.epochs)